/* tclWinidle.c */

#include <tcl.h>
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include "idletracker.h"

#define VERSION "0.2"

static FARPROC GetLastInputInfoF = NULL;
static FARPROC IdleTrackerInitF = NULL;
static FARPROC IdleTrackerGetLastTickCountF = NULL;

/**
 * DLL's entry point
 **/
int WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
{
	return TRUE;
}

/**
 * This is to prevent the CRT from loading, thus making this a smaller
 * and faster dll.
 **/
BOOL __stdcall _DllMainCRTStartup( HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved) {
    return DllMain( hinstDLL, fdwReason, lpvReserved );
}

/*
   tclWinidle::idletime
 */

static int
tclWinidle_idletime( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
    DWORD ticks, lastticks; 
    LASTINPUTINFO lastInputInfo;

    if (objc != 1) {
        Tcl_WrongNumArgs (interp, 1, objv, "");
        return TCL_ERROR;
    }

    ticks = 0;

    if (GetLastInputInfoF == NULL) {
	if (IdleTrackerGetLastTickCountF != NULL)
	    ticks = GetTickCount() - IdleTrackerGetLastTickCountF();
    } else {
	lastInputInfo.cbSize = sizeof(LASTINPUTINFO);
	lastInputInfo.dwTime = 0;

	if (GetLastInputInfoF (&lastInputInfo))
	    ticks = GetTickCount() - lastInputInfo.dwTime;
    }
    if (ticks < 0) ticks = 0;

    Tcl_SetLongObj (Tcl_GetObjResult (interp), ticks);
    return TCL_OK;
}

int
InitIdleTrackerDLL (Tcl_Interp *interp) {
    CONST char *dllname;
    HMODULE hmod;

    dllname = Tcl_GetVar (interp, "::idletracdll", TCL_LEAVE_ERR_MSG);
    if (dllname == NULL) {
	Tcl_AppendResult (interp,
	    " and could not get value of ::idletracdll",
	    (char*)NULL);
	return 0;
    } else {
	hmod = LoadLibrary (dllname);
	if (hmod == NULL) {
	    Tcl_AppendResult (interp,
		"and could not load ", dllname,
		(char*)NULL);
	    return 0;
	} else {
	    IdleTrackerInitF = GetProcAddress (hmod, "IdleTrackerInit");
	    if (IdleTrackerInitF == NULL) {
		Tcl_AppendResult (interp,
		    "and could not get address of IdleTrackerInit",
		    (char*)NULL);
		return 0;
	    }
	    IdleTrackerGetLastTickCountF = GetProcAddress (hmod, "IdleTrackerGetLastTickCount");
	    if (IdleTrackerGetLastTickCountF == NULL) {
		Tcl_AppendResult (interp,
		    "and could not get address of IdleTrackerGetLastTickCount",
		    (char*)NULL);
		return 0;
	    }
	    if (!IdleTrackerInitF()) {
		Tcl_AppendResult (interp,
		    "and IdleTrackerInit returned false",
		    (char*)NULL);
		return 0;
	    }
	}
    }
    return 1;
}

__declspec(dllexport)
int
Tclwinidle_Init (Tcl_Interp *interp) {
    HMODULE hmod;
    int res;
    char *msg = NULL;

    if (Tcl_InitStubs (interp, "8.0", 0) == NULL)
        return TCL_ERROR;

    if (Tcl_PkgProvide (interp, "tclWinidle", VERSION) != TCL_OK)
        return TCL_ERROR;

    /* Load GetLastInputInfo function if available */
    hmod = GetModuleHandle ("USER32.DLL");
    if (hmod == NULL)
	hmod = LoadLibrary ("USER32.DLL");
    if (hmod == NULL) {
	Tcl_AppendResult (interp, "Could not Load USER32.DLL", (char*)NULL);
        res = 0;
    } else {
	GetLastInputInfoF = GetProcAddress (hmod, "GetLastInputInfo");
	if (GetLastInputInfoF == NULL) {
	    Tcl_AppendResult (interp,
		"Could not get address of GetLastInputInfo",
		(char*)NULL);
	    res = 0;
	} else {
	    res = 1;
	}
    }

    /* Load idletracker.dll if GetLastInputInfo is unavailable */
    if (!res && !InitIdleTrackerDLL (interp))
	return TCL_ERROR;

    Tcl_CreateObjCommand (interp, "tclWinidle::idletime", tclWinidle_idletime,
			  (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

    Tcl_ResetResult (interp);
    return TCL_OK;
}
